use crate::db::DBConnection;
use crate::schema::entity::helpers::check_assoc;
use crate::schema::{AssocData, IDWrap, LocalSide};
use crate::{
    schema::datum::{Datum, DatumList, DatumVisitor},
    schema::entity::{Entity, EntityID, EntityPart, EntityPartList, EntityPartVisitor},
    schema::{EntityMap, IDMap},
};
use crate::{DBError, DBResult};
use std::hash::{Hash, Hasher};

#[derive(Hash)]
enum QueryType<'a> {
    ByID(&'a str),
    Delete(std::any::TypeId),
    DeleteAssoc(&'a str, &'a str, &'a str),
    DeleteById,
    Select(std::any::TypeId),
    SelectJoin(&'a str, &'a str, &'a str),
    SelectJoinFilter(&'a str, &'a str, &'a str, std::any::TypeId),
    Insert,
    InsertAssoc(&'a str, &'a str, &'a str),
}

fn bind_datum_to<DL: DatumList>(
    stmt: &mut sqlite::Statement,
    dl: &DL,
    start_index: usize,
) -> DBResult<()> {
    struct BindDatum<'a, 'b>(&'a mut sqlite::Statement<'b>, usize);
    impl<'a, 'b> DatumVisitor for BindDatum<'a, 'b> {
        fn visit<ED: Datum>(&mut self, datum: &ED) {
            datum.bind_to(self.0, self.1);
            self.1 += 1
        }
    }
    // note that this indexing starts at 1
    dl.accept(&mut BindDatum(stmt, start_index));

    Ok(())
}

fn query_hash<E: Entity>(qtype: QueryType) -> u64 {
    let mut hasher = std::collections::hash_map::DefaultHasher::new();

    qtype.hash(&mut hasher);
    std::any::TypeId::of::<E>().hash(&mut hasher);

    hasher.finish()
}

pub(crate) fn by_id<ID: EntityID>(
    conn: &DBConnection,
    id: ID,
) -> DBResult<Option<IDWrap<ID::Entity>>> {
    conn.with_prepared(
        query_hash::<ID::Entity>(QueryType::ByID(<ID::Entity>::entity_name())),
        || {
            format!(
                "select * from {} where `id` = ?",
                <ID::Entity>::entity_name()
            )
        },
        |stmt| {
            stmt.reset()?;
            stmt.bind((1, id.into_raw()))?;

            // now we grab the statement output
            if stmt.next()? == sqlite::State::Row {
                // read the ID column
                let id = stmt.read::<i64, _>(0)?;
                let datum_list = <<ID::Entity as Entity>::Parts>::build_datum_list(conn, stmt)?;
                Ok(Some(IDWrap::new(
                    ID::from_raw(id),
                    <ID::Entity>::build(datum_list),
                )))
            } else {
                Ok(None)
            }
        },
    )
}

pub(crate) fn get_all<E: Entity>(conn: &DBConnection) -> DBResult<Vec<IDWrap<E>>> {
    conn.with_prepared(
        query_hash::<E>(QueryType::Select(std::any::TypeId::of::<E>())),
        || format!("select * from `{}`", E::entity_name()),
        |stmt| {
            stmt.reset()?;

            let mut rows = vec![];
            while stmt.next()? == sqlite::State::Row {
                let datum_list = <E::Parts>::build_datum_list(&conn, stmt)?;
                rows.push(IDWrap::new(
                    <E::ID>::from_raw(stmt.read::<i64, _>(0)?),
                    E::build(datum_list),
                ));
            }

            stmt.reset()?;

            Ok(rows)
        },
    )
}

pub(crate) fn select_by<E: Entity, PL: EntityPartList>(
    map: &IDMap<E>,
    by: &PL::DatumList,
) -> DBResult<Vec<IDWrap<E>>> {
    map.conn().with_prepared(
        query_hash::<E>(QueryType::Select(std::any::TypeId::of::<PL>())),
        || {
            let mut conditions = String::new();
            struct BuildConditions<'a>(&'a mut String);
            impl<'a> EntityPartVisitor for BuildConditions<'a> {
                fn visit<EP: EntityPart>(&mut self) {
                    if self.0.len() > 0 {
                        self.0.push_str(" and ");
                    }
                    self.0.push_str("`");
                    self.0.push_str(EP::part_name());
                    self.0.push_str("`");
                    self.0.push_str(" = ?");
                }
            }
            PL::accept_part_visitor(&mut BuildConditions(&mut conditions));

            let table_name = format!("{}", E::entity_name());
            format!("select * from `{}` where {}", table_name, conditions)
        },
        |stmt| {
            struct BindDatum<'a, 'b>(&'a mut sqlite::Statement<'b>, usize);
            impl<'a, 'b> DatumVisitor for BindDatum<'a, 'b> {
                fn visit<ED: Datum>(&mut self, datum: &ED) {
                    datum.bind_to(self.0, self.1);
                    self.1 += 1
                }
            }
            // note that this indexing starts at 1
            by.accept(&mut BindDatum(stmt, 1));

            // now we grab the statement outputs
            let mut rows = vec![];
            while stmt.next()? == sqlite::State::Row {
                let datum_list = <E::Parts>::build_datum_list(&map.conn(), stmt)?;
                rows.push(IDWrap::new(
                    <E::ID>::from_raw(stmt.read::<i64, _>(0)?),
                    E::build(datum_list),
                ));
            }

            stmt.reset()?;

            Ok(rows)
        },
    )
}

pub(crate) fn delete_by_id<E: Entity>(conn: &DBConnection, id: E::ID) -> DBResult<()> {
    conn.with_prepared(
        query_hash::<E>(QueryType::DeleteById),
        || format!("delete from `{}` where `id` = ?", E::entity_name()),
        |stmt| {
            stmt.reset()?;
            stmt.bind((1, id.into_raw()))?;
            stmt.next()?;
            Ok(())
        },
    )
}

pub(crate) fn delete_by<E: Entity, PL: EntityPartList>(
    conn: &DBConnection,
    values: &PL::DatumList,
) -> DBResult<()> {
    conn.with_prepared(
        query_hash::<E>(QueryType::Delete(std::any::TypeId::of::<PL>())),
        || {
            let mut conditions = String::new();
            struct BuildConditions<'a>(&'a mut String);
            impl<'a> EntityPartVisitor for BuildConditions<'a> {
                fn visit<EP: EntityPart>(&mut self) {
                    if self.0.len() > 0 {
                        self.0.push_str(" and ");
                    }
                    self.0.push_str("`");
                    self.0.push_str(EP::part_name());
                    self.0.push_str("`");
                    self.0.push_str(" = ?");
                }
            }
            PL::accept_part_visitor(&mut BuildConditions(&mut conditions));

            format!("delete from `{}` where {}", E::entity_name(), conditions)
        },
        |stmt| {
            stmt.reset()?;
            bind_datum_to(stmt, values, 1)?;
            stmt.next()?;
            Ok(())
        },
    )
}

pub(crate) fn insert<E: Entity>(conn: &DBConnection, value: E) -> DBResult<E::ID> {
    conn.with_prepared(
        query_hash::<E>(QueryType::Insert),
        || {
            let table_name = format!("{}", E::entity_name());

            let mut part_names = String::new();
            let mut placeholders = String::new();
            struct PartNameVisitor<'a>(&'a mut String, &'a mut String);
            impl<'a> EntityPartVisitor for PartNameVisitor<'a> {
                fn visit<EP: EntityPart>(&mut self) {
                    // if this is a set-association, then we don't actually want to do anything
                    // with it here; it doesn't have a column
                    if check_assoc::<EP>() {
                        return;
                    }

                    if self.0.len() != 0 {
                        self.0.push_str(", ");
                        self.1.push_str(", ");
                    }
                    self.0.push_str("`");
                    self.0.push_str(EP::part_name());
                    self.0.push_str("`");
                    self.1.push_str("?");
                }
            }

            E::accept_part_visitor(&mut PartNameVisitor(&mut part_names, &mut placeholders));

            format!(
                "insert into `{}` ({}) values ({}) returning `id`",
                table_name, part_names, placeholders
            )
        },
        |prepared| {
            prepared.reset()?;

            struct PartBinder<'a, 'b>(&'a mut sqlite::Statement<'b>, usize);
            impl<'a, 'b> EntityPartVisitor for PartBinder<'a, 'b> {
                fn visit_datum<EP: EntityPart>(&mut self, datum: &EP::Datum) {
                    // skip associations, as with the query preparation above
                    if check_assoc::<EP>() {
                        return;
                    }

                    datum.bind_to(self.0, self.1);
                    self.1 += 1;
                }
            }

            value.accept_part_visitor_ref(&mut PartBinder(prepared, 1));

            prepared.next()?;

            Ok(<E::ID>::from_raw(prepared.read::<i64, _>(0)?))
        },
    )
}

pub(crate) struct AssocNames {
    local_name: &'static str,
    remote_name: &'static str,
    part_name: &'static str,
    dist_name: &'static str,
    domain_name: &'static str,
    range_name: &'static str,
    local_field: &'static str,
    remote_field: &'static str,
}

impl AssocNames {
    fn collect<AI: AssocInterface>(iface: &AI) -> DBResult<AssocNames> {
        let adata = iface.get_data()?;
        let local_name = adata.local_name;
        let remote_name = <AI::RemoteEntity>::entity_name();
        let part_name = adata.part_name;
        let dist_name = iface.get_distinguishing_name()?;

        let (domain_name, range_name) = match AI::SIDE {
            LocalSide::Domain => (local_name, remote_name),
            LocalSide::Range => (remote_name, local_name),
        };
        let (local_field, remote_field) = match AI::SIDE {
            LocalSide::Domain => ("domain", "range"),
            LocalSide::Range => ("range", "domain"),
        };
        Ok(Self {
            local_name,
            remote_name,
            part_name,
            dist_name,
            domain_name,
            range_name,
            local_field,
            remote_field,
        })
    }

    fn assoc_name(&self) -> String {
        format!(
            "{domain_name}_{range_name}_assoc_{dist_name}",
            domain_name = self.domain_name,
            range_name = self.range_name,
            dist_name = self.dist_name
        )
    }
}

pub(crate) trait AssocInterface {
    type RemoteEntity: Entity;
    fn get_data(&self) -> DBResult<&AssocData>;
    fn get_distinguishing_name(&self) -> DBResult<&'static str>;
    const SIDE: LocalSide;

    fn get_all(&self) -> DBResult<Vec<IDWrap<Self::RemoteEntity>>>
    where
        Self: Sized,
    {
        let adata = self.get_data()?;

        // equivalent SQL:
        // SELECT
        //     `range_table_name`.*
        // FROM
        //     `assoc_table_name`
        // LEFT JOIN `range_table_name` ON `assoc_table_name`.`range` = `range_table_name`.`rowid`
        // WHERE `domain` = domain_id

        let an = AssocNames::collect::<Self>(&self)?;

        adata.conn.with_prepared(
            query_hash::<Self::RemoteEntity>(QueryType::SelectJoin(
                an.local_name,
                an.remote_name,
                an.part_name,
            )),
            || {
                format!(
                    "select `{remote_name}`.* from `{assoc_name}` \
                        left join `{remote_name}` on \
                            `{assoc_name}`.`{remote_field}` = `{remote_name}`.`id` \
                        where `{assoc_name}`.`{local_field}` = ?",
                    assoc_name = an.assoc_name(),
                    remote_name = an.remote_name,
                    local_field = an.local_field,
                    remote_field = an.remote_field,
                )
            },
            |stmt| {
                stmt.bind((1, adata.local_id))?;

                // now we grab the statement outputs
                let mut rows = vec![];
                while stmt.next()? == sqlite::State::Row {
                    let id = stmt.read::<i64, _>(0)?;
                    let datum_list = <<Self::RemoteEntity as Entity>::Parts>::build_datum_list(
                        &adata.conn,
                        stmt,
                    )?;
                    rows.push(IDWrap::new(
                        <Self::RemoteEntity as Entity>::ID::from_raw(id),
                        Self::RemoteEntity::build(datum_list),
                    ));
                }

                Ok(rows)
            },
        )
    }

    fn lookup_by<PL: EntityPartList>(
        &self,
        values: &PL::DatumList,
    ) -> DBResult<Vec<IDWrap<Self::RemoteEntity>>>
    where
        Self: Sized,
    {
        let adata = self.get_data()?;

        // equivalent SQL:
        // SELECT
        //     `range_table_name`.*
        // FROM
        //     `assoc_table_name`
        // LEFT JOIN `range_table_name` ON `assoc_table_name`.`range` = `range_table_name`.`rowid`
        // WHERE
        //      `domain` = domain_id
        //      AND `unique_1` = ?

        let an = AssocNames::collect::<Self>(&self)?;

        adata.conn.with_prepared(
            query_hash::<Self::RemoteEntity>(QueryType::SelectJoinFilter(
                an.local_name,
                an.remote_name,
                an.part_name,
                std::any::TypeId::of::<<Self::RemoteEntity as Entity>::Uniques>(),
            )),
            || {
                let mut conditions = String::new();
                struct BuildConditions<'a>(&'a mut String);
                impl<'a> EntityPartVisitor for BuildConditions<'a> {
                    fn visit<EP: EntityPart>(&mut self) {
                        self.0.push_str(" and ");
                        self.0.push_str("`");
                        self.0.push_str(EP::part_name());
                        self.0.push_str("`");
                        self.0.push_str(" = ?");
                    }
                }

                <<Self::RemoteEntity as Entity>::Uniques>::accept_part_visitor(
                    &mut BuildConditions(&mut conditions),
                );

                format!(
                    "select `{remote_name}`.* from `{assoc_name}` \
                        left join `{remote_name}` on \
                            `{assoc_name}`.`{remote_field}` = `{remote_name}`.`id` \
                        where `{assoc_name}`.`{local_field}` = ? {conditions}",
                    assoc_name = an.assoc_name(),
                    remote_name = an.remote_name,
                    local_field = an.local_field,
                    remote_field = an.remote_field,
                )
            },
            |stmt| {
                stmt.bind((1, adata.local_id))?;
                bind_datum_to(stmt, values, 2);

                // now we grab the statement outputs
                let mut rows = vec![];
                while stmt.next()? == sqlite::State::Row {
                    let id = stmt.read::<i64, _>(0)?;
                    let datum_list = <<Self::RemoteEntity as Entity>::Parts>::build_datum_list(
                        &adata.conn,
                        stmt,
                    )?;
                    rows.push(IDWrap::new(
                        <Self::RemoteEntity as Entity>::ID::from_raw(id),
                        Self::RemoteEntity::build(datum_list),
                    ));
                }

                Ok(rows)
            },
        )
    }

    fn associate_with(&self, remote_id: <Self::RemoteEntity as Entity>::ID) -> DBResult<()>
    where
        Self: Sized,
    {
        let adata = self.get_data()?;
        let an = AssocNames::collect::<Self>(&self)?;

        // second, add to the assoc table
        adata.conn.with_prepared(
            query_hash::<Self::RemoteEntity>(QueryType::InsertAssoc(
                an.local_name,
                an.remote_name,
                an.part_name,
            )),
            || {
                format!(
                    "insert into `{assoc_name}` (`{local_field}`, `{remote_field}`) values (?, ?)",
                    assoc_name = an.assoc_name(),
                    local_field = an.local_field,
                    remote_field = an.remote_field
                )
            },
            |stmt| {
                stmt.reset()?;
                stmt.bind((1, adata.local_id))?;
                stmt.bind((2, remote_id.into_raw()))?;

                stmt.next()?;

                Ok(())
            },
        )
    }

    fn dissociate_with(&self, remote_id: <Self::RemoteEntity as Entity>::ID) -> DBResult<()>
    where
        Self: Sized,
    {
        let adata = self.get_data()?;
        let an = AssocNames::collect::<Self>(&self)?;

        // second, add to the assoc table
        adata.conn.with_prepared(
            query_hash::<Self::RemoteEntity>(QueryType::DeleteAssoc(
                an.local_name,
                an.remote_name,
                an.part_name,
            )),
            || {
                format!(
                    "delete from `{assoc_name}` where `{local_field}` = ? and `{remote_field}` = ?",
                    assoc_name = an.assoc_name(),
                    local_field = an.local_field,
                    remote_field = an.remote_field
                )
            },
            |stmt| {
                stmt.reset()?;
                stmt.bind((1, adata.local_id))?;
                stmt.bind((2, remote_id.into_raw()))?;

                stmt.next()?;

                Ok(())
            },
        )
    }

    fn insert(&self, value: Self::RemoteEntity) -> DBResult<<Self::RemoteEntity as Entity>::ID>
    where
        Self: Sized,
    {
        // we're doing two things:
        // - inserting the entity into the target table
        // - adding the association row into the assoc table

        let adata = self.get_data()?;

        // so first, the remote table
        let remote_id = insert(&adata.conn, value)?;
        // then the association
        self.associate_with(remote_id)?;
        // TODO: handle error case of associate_with() fails but insert() succeeds
        Ok(remote_id)
    }
}
