#![allow(unused)]

fn open_test_db<DB: super::Database>(identifier: &'static str) -> DB {
    let path = format!("/tmp/microrm-{identifier}.db");
    let _ = std::fs::remove_file(path.as_str());
    DB::open_path(path).expect("couldn't open database file")
}

mod manual_test_db {
    // simple hand-built database example

    use crate::db::DBConnection;
    use crate::schema::datum::Datum;
    use crate::schema::entity::{
        Entity, EntityID, EntityPart, EntityPartList, EntityPartVisitor, EntityVisitor,
    };
    use crate::schema::{Database, DatabaseItem, DatabaseItemVisitor, IDMap};

    struct SimpleEntity {
        name: String,
    }

    #[derive(Clone, Copy, PartialEq, PartialOrd, Debug, Hash)]
    struct SimpleEntityID(i64);

    impl Datum for SimpleEntityID {
        fn sql_type() -> &'static str {
            "int"
        }
        fn accept_entity_visitor(_: &mut impl EntityVisitor) {}
        fn accept_discriminator(d: &mut impl crate::schema::DatumDiscriminator)
        where
            Self: Sized,
        {
            d.visit_entity_id::<<Self as EntityID>::Entity>();
        }

        fn bind_to<'a>(&self, _stmt: &mut sqlite::Statement<'a>, index: usize) {
            todo!()
        }
        fn build_from<'a>(
            adata: crate::schema::AssocData,
            stmt: &mut sqlite::Statement<'a>,
            index: usize,
        ) -> crate::DBResult<(Self, usize)>
        where
            Self: Sized,
        {
            todo!()
        }
    }

    impl EntityID for SimpleEntityID {
        type Entity = SimpleEntity;

        fn from_raw(id: i64) -> Self {
            Self(id)
        }
        fn into_raw(self) -> i64 {
            self.0
        }
    }

    struct SimpleEntityName;
    impl EntityPart for SimpleEntityName {
        type Datum = String;
        type Entity = SimpleEntity;
        fn part_name() -> &'static str {
            "name"
        }
        fn unique() -> bool {
            true
        }
    }

    impl Entity for SimpleEntity {
        type Parts = SimpleEntityName;
        type Uniques = SimpleEntityName;
        type ID = SimpleEntityID;

        fn build(name: String) -> Self {
            Self { name }
        }

        fn entity_name() -> &'static str {
            "simple_entity"
        }
        fn accept_part_visitor(visitor: &mut impl EntityPartVisitor) {
            visitor.visit::<SimpleEntityName>();
        }
        fn accept_part_visitor_ref(&self, visitor: &mut impl EntityPartVisitor) {
            visitor.visit_datum::<SimpleEntityName>(&self.name);
        }
    }

    struct SimpleDatabase {
        strings: IDMap<SimpleEntity>,
    }

    impl Database for SimpleDatabase {
        fn build(conn: DBConnection) -> Self
        where
            Self: Sized,
        {
            Self {
                strings: IDMap::build(conn),
            }
        }

        fn accept_item_visitor(visitor: &mut impl DatabaseItemVisitor)
        where
            Self: Sized,
        {
            struct SimpleDatabaseStringsItem;
            impl DatabaseItem for SimpleDatabaseStringsItem {
                fn item_key() -> &'static str {
                    "strings"
                }
                fn dependency_keys() -> &'static [&'static str] {
                    &[]
                }

                fn accept_entity_visitor(visitor: &mut impl EntityVisitor) {
                    visitor.visit::<SimpleEntity>();
                }
            }

            visitor.visit::<SimpleDatabaseStringsItem>();
        }
    }

    #[test]
    fn part_visitor() {
        struct V {
            v: Vec<std::any::TypeId>,
        }
        impl EntityPartVisitor for V {
            fn visit<EP: EntityPart>(&mut self) {
                self.v.push(std::any::TypeId::of::<EP>());
            }
        }

        let mut vis = V { v: vec![] };
        SimpleEntity::accept_part_visitor(&mut vis);
        assert_eq!(
            vis.v.as_slice(),
            &[std::any::TypeId::of::<SimpleEntityName>()]
        );
    }
}

mod derive_tests {
    #![allow(unused)]

    use crate::query::AssocInterface;
    use crate::schema::{AssocMap, Database, IDMap};
    use microrm_macros::{Database, Entity};

    #[derive(Entity, Debug)]
    struct Role {
        title: String,
        permissions: String,
    }

    #[derive(Entity, Debug)]
    struct Person {
        #[unique]
        name: String,
        roles: AssocMap<Role>,
    }

    #[derive(Database)]
    struct PeopleDB {
        people: IDMap<Person>,
    }

    #[test]
    fn collect_test() {
        microrm::schema::build::collect_from_database::<PeopleDB>();
    }

    #[test]
    fn open_test() {
        PeopleDB::open_path(":memory:");
    }

    #[test]
    fn check_for_inserted() {
        std::fs::remove_file("/tmp/schema.db").ok();
        let db = PeopleDB::open_path("/tmp/schema.db").expect("couldn't open database");
        // let db = PeopleDB::open_path(":memory:").expect("couldn't open database");
        let name_string = "name_here".to_string();
        // check that it isn't in the database before we insert it
        assert!(db
            .people
            .lookup_unique(&name_string)
            .ok()
            .flatten()
            .is_none());

        db.people
            .insert(Person {
                name: name_string.clone(),
                roles: AssocMap::empty(),
            })
            .expect("failed to insert");

        // check that it is in the database after we insert it
        assert!(db
            .people
            .lookup_unique(&name_string)
            .ok()
            .flatten()
            .is_some());
    }

    #[test]
    fn check_assoc_query_construction() {
        let db = PeopleDB::open_path(":memory:").expect("couldn't open database");

        let name_string = "name_here".to_string();
        db.people
            .insert(Person {
                name: name_string.clone(),
                roles: AssocMap::empty(),
            })
            .expect("couldn't insert test person");

        let person = db
            .people
            .lookup_unique(&name_string)
            .ok()
            .flatten()
            .expect("couldn't re-get test person entity");

        person.roles.get_all();
    }

    #[test]
    fn check_assoc_insertion() {
        std::fs::remove_file("/tmp/insert.db").ok();
        let db = PeopleDB::open_path("/tmp/insert.db").expect("couldn't open database");
        // let db = PeopleDB::open_path(":memory:").expect("couldn't open database");

        let name_string = "name_here".to_string();
        db.people
            .insert(Person {
                name: name_string.clone(),
                roles: AssocMap::empty(),
            })
            .expect("couldn't insert test person");

        let person = db
            .people
            .lookup_unique(&name_string)
            .ok()
            .flatten()
            .expect("couldn't re-get test person entity");

        person.roles.insert(Role {
            title: "title A".to_string(),
            permissions: "permissions A".to_string(),
        });
    }
}

mod mutual_relationship {
    use super::open_test_db;
    use crate::query::AssocInterface;
    use crate::schema::{AssocDomain, AssocMap, AssocRange, Database, IDMap};
    use microrm_macros::{Database, Entity};

    struct CR;
    impl microrm::schema::Relation for CR {
        type Domain = Customer;
        type Range = Receipt;

        const NAME: &'static str = "CR";
    }

    #[derive(Entity)]
    struct Customer {
        name: String,
        receipts: AssocDomain<CR>,
    }

    #[derive(Entity)]
    struct Receipt {
        value: usize,
        customers: AssocRange<CR>,
    }

    #[derive(Database)]
    struct ReceiptDB {
        customers: IDMap<Customer>,
        receipts: IDMap<Receipt>,
    }

    #[test]
    fn check_schema_creation() {
        let db = open_test_db::<ReceiptDB>("mutual_relationship_create");
    }

    #[test]
    fn simple_operations() {
        let db = open_test_db::<ReceiptDB>("mutual_relationship_simple");

        let ca = db
            .customers
            .insert(Customer {
                name: "customer A".to_string(),
                receipts: Default::default(),
            })
            .expect("couldn't insert customer record");

        let ra = db
            .receipts
            .insert(Receipt {
                value: 32usize,
                customers: Default::default(),
            })
            .expect("couldn't insert receipt record");

        let rb = db
            .receipts
            .insert(Receipt {
                value: 64usize,
                customers: Default::default(),
            })
            .expect("couldn't insert receipt record");

        let e_ca = db
            .customers
            .by_id(ca)
            .expect("couldn't retrieve customer record")
            .expect("no customer record");
        e_ca.receipts
            .associate_with(ra)
            .expect("couldn't associate customer with receipt");
        e_ca.receipts
            .associate_with(rb)
            .expect("couldn't associate customer with receipt");

        // technically this can fail if sqlite gives ra and rb back in the opposite order, which is
        // valid behaviour
        assert_eq!(
            e_ca.receipts
                .get_all()
                .expect("couldn't get receipts associated with customer")
                .into_iter()
                .map(|x| x.id())
                .collect::<Vec<_>>(),
            vec![ra, rb]
        );

        // check that the reverse direction was also added
        let e_ra = db
            .receipts
            .by_id(ra)
            .expect("couldn't retreieve receipt record")
            .expect("no receipt record");

        assert_eq!(
            e_ra.customers
                .get_all()
                .expect("couldn't get associated customers")
                .into_iter()
                .map(|x| x.id())
                .collect::<Vec<_>>(),
            vec![ca]
        );
    }
}

mod reserved_words {
    use crate::db::DBConnection;
    use crate::prelude::*;
    use crate::schema::entity::Entity;
    use crate::schema::{AssocDomain, AssocRange, Database, IDMap};

    #[derive(Entity)]
    struct Select {
        delete: String,
    }

    #[derive(Entity)]
    struct Group {
        by: AssocMap<Select>,
    }

    #[derive(Database)]
    struct ReservedWordDB {
        group: IDMap<Group>,
    }

    #[test]
    fn open_test() {
        ReservedWordDB::open_path(":memory:");
    }
}
